/*    %W% %G%
================================================================================
  Description :
  Author      : Ryu, Byeong-Soon
================================================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include "socket.h"
#include "pxhan.h"

/*#define TRACE_ON*/
#include "debug.h"

static Channel  *_Channels;
static int  _NumChannels = 0;

int  VXSockFD;
extern char VXByteOrder;
extern void VXGetAddress(char *, int *);

void
VXInitChannel()
{
	_Channels = NULL;
	_NumChannels = 0;
}

void
VXNewChannel()
{
	int  port, cfd, sfd;
	char  addr[100];
	Channel  *new_cl;
	int  len;
	struct sockaddr_in  from;

	TRACE("new channel\n");
	len = sizeof(from);
	getsockname(VXSockFD, (struct sockaddr *) &from, &len);

	cfd = accept(VXSockFD, (struct sockaddr *) &from, &len);
	if(cfd < 0) {
		perror("accept");
		return;
	}

	VXGetAddress(addr, &port);
	TRACE2("connect to X-server: %s, %d\n", addr, port);
	sfd = SockInitializeClient(addr, port);
	if(sfd < 0) {
		return;
	}

	new_cl = (Channel *) malloc(sizeof(Channel));
	new_cl->next = _Channels;
	_Channels = new_cl;
	_NumChannels++;

	new_cl->cfd = cfd;
	new_cl->sfd = sfd;

	/* clear new channel */
	new_cl->c_connected = 0;
	new_cl->s_connected = 0;
	new_cl->ci_buffer = (char*)malloc(MAX_BUFFER);
	new_cl->ci_buff_size = 0;
	new_cl->ci_buff_max = MAX_BUFFER;
	new_cl->co_buffer = (char*)malloc(MAX_BUFFER);
	new_cl->co_buff_size = 0;
	new_cl->co_buff_max = MAX_BUFFER;
	new_cl->si_buffer = (char*)malloc(MAX_BUFFER);
	new_cl->si_buff_size = 0;
	new_cl->si_buff_max = MAX_BUFFER;
	new_cl->so_buffer = (char*)malloc(MAX_BUFFER);
	new_cl->so_buff_size = 0;
	new_cl->so_buff_max = MAX_BUFFER;
	new_cl->c_req_len = 0;
	new_cl->s_req_len = 0;
	new_cl->ctemp = (char*)malloc(MAX_BUFFER);
	new_cl->c_temp_size = MAX_BUFFER;
	new_cl->c_str_req_len = 0;
	new_cl->c_str_len = 0;
	new_cl->c_ipos = 0;
	new_cl->c_opos = 0;
	new_cl->s_ipos = 0;
	new_cl->s_opos = 0;
	new_cl->c_len_field = -1;
	new_cl->hangul = 0;
	new_cl->input_temp = 0;
	new_cl->font_loaded = 0;
	new_cl->font_gc = NULL;
}

void
VXFreeChannel(Channel *pre, Channel *ptr)
{
	TRACE("close channel\n");
	if(pre == NULL) {
		_Channels = ptr->next;
	} else {
		pre->next = ptr->next;
	}

	close(ptr->cfd);
	close(ptr->sfd);
	free(ptr->ci_buffer);
	free(ptr->co_buffer);
	free(ptr->si_buffer);
	free(ptr->so_buffer);
	if (ptr->ctemp)
		free(ptr->ctemp);
	free(ptr);
	_NumChannels--;
}

void
hupsig()
{
	Channel  *ptr;

	puts("signal detected...");

	for(ptr = _Channels; ptr != NULL; ptr = ptr->next) {
		close(ptr->cfd);
		close(ptr->sfd);
		free(ptr->ci_buffer);
		free(ptr->co_buffer);
		free(ptr->si_buffer);
		free(ptr->so_buffer);
		if (ptr->ctemp)
			free(ptr->ctemp);
	}
	close(VXSockFD);
	exit(0);
}

void
VXMainLoop()
{
	int  max_desc = ulimit(4, 0);
	fd_set  in_set, out_set, exc_set;
	register Channel  *ptr;

	TRACE1("max_desc : %d\n", max_desc);

	signal(SIGHUP, hupsig);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGINT, hupsig);
	signal(SIGTERM, hupsig);

	TRACE("main loop\n");
	while(1) {
		FD_ZERO(&in_set);
		FD_ZERO(&out_set);
		FD_ZERO(&exc_set);

		FD_SET(VXSockFD, &in_set);
		for(ptr = _Channels; ptr != NULL; ptr = ptr->next) {
			/* exceptions */
			FD_SET(ptr->sfd, &exc_set);
			FD_SET(ptr->cfd, &exc_set);

			/* outputs */
			if( ptr->so_buff_size > 0 ||
				ptr->ci_buff_size > 0 ||
				ptr->input_temp == 1
			) {
				FD_SET(ptr->sfd, &out_set);
			}
			if(ptr->co_buff_size > 0 || ptr->si_buff_size > 0)
				FD_SET(ptr->cfd, &out_set);

			/* inputs */
			if(ptr->si_buff_size < MAX_BUFFER)
				FD_SET(ptr->sfd, &in_set);
			if(ptr->ci_buff_size < MAX_BUFFER)
				FD_SET(ptr->cfd, &in_set);
		}

		if(select(max_desc, &in_set, &out_set, &exc_set, NULL) < 0) {
			perror("select");
			return;
		}

		VXProcessInput(&in_set, &out_set, &exc_set);
	}
}

void
VXProcessInput(fd_set *in_set, fd_set *out_set, fd_set *exc_set)
{
	register Channel  *ptr, *old_ptr, *next;

	/* New connection? */
	if(FD_ISSET(VXSockFD, in_set)) {
		VXNewChannel();
	}

	/* Kick out the freaky folks */
	old_ptr = NULL;
	for(ptr = _Channels; ptr != NULL; ptr = next) {
		next = ptr->next;
		if(FD_ISSET(ptr->cfd, exc_set) || FD_ISSET(ptr->sfd, exc_set)) {
			FD_CLR(ptr->cfd, in_set);
			FD_CLR(ptr->cfd, out_set);
			FD_CLR(ptr->sfd, in_set);
			FD_CLR(ptr->sfd, out_set);
			VXFreeChannel(old_ptr, ptr);
		} else {
			old_ptr = ptr;
		}
	}

	old_ptr = NULL;
	for(ptr = _Channels; ptr != NULL; ptr = next) {
		next = ptr->next;

		if(FD_ISSET(ptr->sfd, in_set)) {
			if(VXInputServer(ptr) <= 0) {
				VXFreeChannel(old_ptr, ptr);
				ptr = old_ptr;
				continue;
			}
		}
		if(FD_ISSET(ptr->cfd, out_set)) {
			if(VXOutputClient(ptr) < 0) {
				VXFreeChannel(old_ptr, ptr);
				continue;
			}
		}

		if(FD_ISSET(ptr->cfd, in_set)) {
			if(VXInputClient(ptr) <= 0) {
				VXFreeChannel(old_ptr, ptr);
				ptr = old_ptr;
				continue;
			}
		}
		if(FD_ISSET(ptr->sfd, out_set)) {
			if(VXOutputServer(ptr) < 0) {
				VXFreeChannel(old_ptr, ptr);
				continue;
			}
		}
		old_ptr = ptr;
	}
}

int
VXInputClient(Channel *ptr)
{
	int  nleft, nread;

	nleft = MAX_BUFFER - ptr->ci_buff_size;
	if(nleft == 0)
		return 0;

	nread = read(ptr->cfd, ptr->ci_buffer+ptr->ci_buff_size, nleft);
	if(nread < 0)
		return nread;
	TRACE1("input client: nread = %d\n", nread);

	ptr->ci_buff_size += nread;

	return nread;
}

int
VXInputServer(Channel *ptr)
{
	int  nleft, nread;

	nleft = MAX_BUFFER - ptr->si_buff_size;
	if(nleft == 0)
		return 0;

	nread = read(ptr->sfd, ptr->si_buffer+ptr->si_buff_size, nleft);
	if(nread < 0)
		return nread;
	TRACE1("input server: nread = %d\n", nread);

	ptr->si_buff_size += nread;

	return nread;
}

int
VXOutputClient(Channel *ptr)
{
	int  nwritten;

	VXFilterServerInput(ptr);
	if(ptr->co_buff_size == 0)
		return 0;

	nwritten = write(ptr->cfd, ptr->co_buffer, ptr->co_buff_size);
	if(nwritten < 0)
		return nwritten;
	TRACE2("output client: size = %d, nwritten = %d\n",
		ptr->co_buff_size, nwritten
	);

	ptr->co_buff_size -= nwritten;
	if(ptr->co_buff_size > 0)
		memmove(ptr->co_buffer, ptr->co_buffer+nwritten, ptr->co_buff_size);
	return  nwritten;
}

int
VXOutputServer(Channel *ptr)
{
	int  nwritten;

	VXFilterClientInput(ptr);
	if(ptr->so_buff_size == 0)
		return 0;

	nwritten = write(ptr->sfd, ptr->so_buffer, ptr->so_buff_size);
	if(nwritten < 0)
		return nwritten;

	ptr->so_buff_size -= nwritten;
	if(ptr->so_buff_size > 0)
		memmove(ptr->so_buffer, ptr->so_buffer+nwritten, ptr->so_buff_size);
	return  nwritten;
}
