#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "pxhan.h"

/*#define TRACE_ON*/
#include "debug.h"

/* number of additional requests : for hangul font */
#define ADDITIONAL_REQUESTS		2

extern char VXByteOrder;

void
VXMemCopyClient(Channel *xptr, unsigned long len)
{
	register Channel *ptr;
	register char *temp;
	ptr = xptr;
	if (ptr->so_buff_size + len > ptr->so_buff_max)
	{
		TRACE("VXMemCopyClient: so_buff overflow enlarging..\n");
		while (ptr->so_buff_max < ptr->so_buff_size + len)
			ptr->so_buff_max += MAX_BUFFER;
		temp = (char*)realloc(ptr->so_buffer, ptr->so_buff_max);
		if (temp == NULL)
		{
			TRACE("VXMemCopyClient: Out of mem\n");
			exit(4);
		}
		TRACE1("VXMemCopyClient: so_buff enlarged to %d bytes.\n", ptr->so_buff_max);
		ptr->so_buffer = temp;
	}
	memcpy(ptr->so_buffer+ptr->so_buff_size, ptr->ci_buffer, len);
	memmove(ptr->ci_buffer, ptr->ci_buffer+len, ptr->ci_buff_size-len);
	ptr->ci_buff_size -= len;
	ptr->so_buff_size += len;
	if (ptr->c_req_len > len)
		ptr->c_req_len -= len;
	else
		ptr->c_req_len = 0;
}

void
VXMemCopyServer(Channel *xptr, unsigned long len)
{
	register char *temp;
	register Channel *ptr;
	ptr = xptr;
	if (ptr->co_buff_size + len > ptr->co_buff_max)
	{
		TRACE("VXMemCopyServer: co_buff overflow enlarging..\n");
		while (ptr->co_buff_max < ptr->co_buff_size + len)
			ptr->co_buff_max += MAX_BUFFER;
		temp = (char*)realloc(ptr->co_buffer, ptr->co_buff_max);
		if (temp == NULL)
		{
			TRACE("VXMemCopyServer: Out of mem\n");
			exit(4);
		}
		TRACE1("VXMemCopyServer: co_buff enlarged to %d bytes\n", ptr->co_buff_max);
		ptr->co_buffer = temp;
	}
	memcpy(ptr->co_buffer+ptr->co_buff_size, ptr->si_buffer, len);
	memmove(ptr->si_buffer, ptr->si_buffer+len, ptr->si_buff_size-len);
	ptr->si_buff_size -= len;
	ptr->co_buff_size += len;
	ptr->s_req_len -= len;
}

void
VXFilterClientInput(Channel *xptr)
{
	register Channel *ptr;
	register long  len, dlen;
	register char *temp;

	ptr = xptr;
	if(ptr->input_temp) {
		TRACE("input_temp\n");
		/* len = min(available space, number of data to be output) */
		len = ptr->c_opos;
		if (ptr->so_buff_size + len > ptr->so_buff_max)
		{
			TRACE("VXFilterClientInput: so_buff overflow enlarging..\n");
			while (ptr->so_buff_max < ptr->so_buff_size + len)
				ptr->so_buff_max += MAX_BUFFER;
			temp = (char*)realloc(ptr->so_buffer, ptr->so_buff_max);
			if (temp == NULL)
			{
				TRACE("VXMemCopyClient: Out of mem\n");
				exit(4);
			}
			TRACE1("VXFilterClientInput: so_buff enlarged to %d bytes\n", ptr->so_buff_max);
			ptr->so_buffer = temp;
		}
		memcpy(ptr->so_buffer+ptr->so_buff_size, ptr->ctemp, len);
		memmove(ptr->ctemp, ptr->ctemp+len, ptr->c_opos-len);
		ptr->c_opos -= len;
		ptr->so_buff_size += len;
		if(ptr->c_opos <= 0) {
			ptr->input_temp = 0;
			ptr->c_req_len = 0;
		}
		return;
	}

	/* len = min(available space, number of data to be output) */
	len = ptr->ci_buff_size;
	while(len > 0) {
		if(ptr->c_req_len == 0) {		/* new request */
			if(ptr->s_connected == 1 && ptr->font_loaded == 0) {
				VXInitHangulFont(ptr);
				ptr->font_loaded = 1;
			}
			ptr->c_ipos = 0;
			ptr->c_str_len = 0;
			ptr->opcode = ptr->ci_buffer[0];
			if(ptr->c_connected == 0) {
				if(ptr->opcode == OP_MSB_FIRST || ptr->opcode == OP_LSB_FIRST) {
					ptr->byte_order = ptr->opcode;
					if(ptr->byte_order == OP_MSB_FIRST)
						puts("Byte order of this connection is MSB First");
					else
						puts("Byte order of this connection is LSB First");

					/* length field is not on memory */
					if(ptr->ci_buff_size < 10) {
						TRACE("init: *** NO LENGTH FIELD ***\n");
						break;
					}

					/* len auth-name */
					dlen = VXMemShort(ptr, ptr->ci_buffer, 6);
					/* len auth-data */
					dlen += VXMemShort(ptr, ptr->ci_buffer, 8);
					ptr->c_req_len = dlen * 4 + 12;
				} else {	/* pass error, as it is, to the server */
					VXMemCopyClient(ptr, len);
					len = 0;
				}
			} else {
				/* length field is not on memory */
				if(ptr->ci_buff_size < 4) {
					TRACE("*** NO LENGTH FIELD ***\n");
					break;
				}

				dlen = VXMemShort(ptr, ptr->ci_buffer, 2);
				ptr->c_req_len = dlen * 4;
			}
			TRACE2("OP_CODE = %d, len = %d\n", ptr->opcode, ptr->c_req_len);
		} else {	/* body of request */
			switch(ptr->opcode) {
			case OP_IMAGE_TEXT8:
				if(ptr->c_req_len <= len) {
					dlen = ptr->c_req_len;
					VXFilterImageText8(ptr);
					if(ptr->input_temp)
						len = 0;
					else
						len -= dlen;
				} else {
					len = 0;
				}
				break;
			case OP_POLY_TEXT8:
				VXFilterPolyText8(ptr);
				len = 0;
				break;
			case OP_CREATE_GC:
			case OP_CHANGE_GC:
				/* process after reading entire request */
				if(ptr->c_req_len <= len) {
					VXFilterGC(ptr);
					len -= ptr->c_req_len;
					VXMemCopyClient(ptr, ptr->c_req_len);
				} else
					len = 0;
				break;
			case OP_FREE_GC:
				/* process after reading entire request */
				if(ptr->c_req_len <= len) {
					VXFilterFreeGC(ptr);
					len -= ptr->c_req_len;
					VXMemCopyClient(ptr, ptr->c_req_len);
				} else
					len = 0;
				break;
			case OP_COPY_GC:
				/* process after reading entire request */
				if(ptr->c_req_len <= len) {
					VXFilterCopyGC(ptr);
					len -= ptr->c_req_len;
					VXMemCopyClient(ptr, ptr->c_req_len);
				} else
					len = 0;
				break;
			default:
				/* copy data */
				if(ptr->c_req_len <= len) {
					len -= ptr->c_req_len;
					VXMemCopyClient(ptr, ptr->c_req_len);
					if(ptr->c_connected == 0) {
						ptr->c_connected = 1;
					}
				} else {
					VXMemCopyClient(ptr, len);
					len = 0;
				}
				break;
			}
		}
	}
}

void
VXFilterServerInput(Channel *xptr)
{
	register Channel *ptr;
	register int  len;
	register unsigned short  seq_num;

	ptr = xptr;
	/* len = min(available space, number of data to be output) */
	len = ptr->si_buff_size;
	while(len > 0) {
		if(ptr->s_req_len == 0) {
			if(ptr->s_connected == 0) {
				if(ptr->si_buffer[0] == 1) {
					if(ptr->si_buff_size >= 24) {
						ptr->res_base = VXMemLong(ptr, ptr->si_buffer, 12);
						ptr->res_mask = VXMemLong(ptr, ptr->si_buffer, 16);
						TRACE2("BASE, MASK = %x, %x\n", ptr->res_base, ptr->res_mask);
					} else
						break;
				}
				if(ptr->si_buff_size >= 8) {
					ptr->s_req_len = VXMemShort(ptr, ptr->si_buffer, 6) * 4;
					ptr->s_req_len += 8;
					ptr->s_connected = 1;
				}
			} else {
				if(ptr->si_buff_size >= 8) {
					seq_num = VXMemShort(ptr, ptr->si_buffer, 2);
					if(seq_num > ADDITIONAL_REQUESTS) {
						seq_num -= ADDITIONAL_REQUESTS;
						if (ptr->byte_order == VXByteOrder)
						{
							if (VXByteOrder == OP_MSB_FIRST)
								VXSetMemMSBShort(ptr->si_buffer, 2, seq_num);
							else
								VXSetMemLSBShort(ptr->si_buffer, 2, seq_num);
						}
						else
						{
							if (VXByteOrder == OP_MSB_FIRST)
								VXSetMemLSBShort(ptr->si_buffer, 2, seq_num);
							else
								VXSetMemMSBShort(ptr->si_buffer, 2, seq_num);
						}
					}
					if(ptr->si_buffer[0] != 1)
						ptr->s_req_len = 32;
					else if(ptr->si_buff_size >= 8)
						ptr->s_req_len = 32 +
							VXMemLong(ptr, ptr->si_buffer, 4)*4;
				}
			}
		}
	
		if(ptr->s_req_len > 0) {
			if(ptr->s_req_len <= len) {
				len -= ptr->s_req_len;
				VXMemCopyServer(ptr, ptr->s_req_len);
			} else {
				VXMemCopyServer(ptr, len);
				len = 0;
			}
		} else
			break;
	}
}
