#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include "socket.h"
#include "pxhan.h"

/*#define TRACE_ON*/
#include "debug.h"

/* default fonts */
#define DEF_HAN_FONT		"-kaist-iyagi-bold-r-normal--16-160-75-75-c-160-johabsh-1"
#define DEF_ASC_FONT		"10x20"

/* environment variables */
#define ENV_HAN_FONT		"PXHAN_HAN_FONT"
#define ENV_ASC_FONT		"PXHAN_ASC_FONT"
#define ENV_HAN_FONT_TYPE	"PXHAN_HAN_FONT_TYPE"

static char  _Address[100];
static int  _Port;

extern int PXgethostname(char *, int);
extern int PXgethostaddr(char *, char *);

void
VXUsage()
{
	fprintf(stderr, "pxhan [-display display] [command list...]\n");
	exit(0);
}
void
VXSetAddress(char *dname)
{
	int  i;

	if(dname == NULL) {
		fprintf(stderr, "DISPLAY is not set !!!\n");
		exit(0);
	}

	strcpy(_Address, dname);
	for(i = 0; i < strlen(dname); i++) {
		if(dname[i] == ':') {
			if(i == 0)
				gethostname(_Address, 100);
			else
				_Address[i] = '\0';
			_Port = XPORT + atoi(&dname[i+1]);
			fprintf(stderr, "SERVER = %s:%d\n", _Address, _Port-XPORT);
			return;
		}
	}

	VXUsage();
}

void
VXGetAddress(char *addr, int *port)
{
	strcpy(addr, _Address);
	*port = _Port;
}

#if defined(SIGCLD) && !defined(SIGCHLD)
# define SIGCHLD	SIGCLD
#endif

#ifdef SIGCHLD
void sig_child()
{
	fprintf(stderr, "pxhan is terminated.\n");
	exit(0);
}
#endif

int  VXLocalPort, VXHanFontType;
char  VXHanFontName[100], VXASCFontName[100];

int main(int argc, char **argv)
{
	int  port, arg_cmd;
	char  hostname[100], display[100], hostaddr[100];
	char  *han_font, *asc_font, *han_font_type, *tmp;
	pid_t	pid1;
#ifndef SIGCHLD
	pid_t   pid2;
#endif
	char buf[100];

	/* get display */
	if(argc >= 3 && strncmp(argv[1], "-disp", 5) == 0) {
		arg_cmd = 3;
		VXSetAddress(argv[2]);
	} else {
		arg_cmd = 1;
		VXSetAddress(getenv("DISPLAY"));
	}
	VXGetAddress(hostname, &port);
	sprintf(display, "%s:%d.0\n", hostname, port-XPORT);

	/* get port */
	for(port = XPORT+1; (VXSockFD = SockInitializeServer(port)) < 0;port++);
	PXgethostname(hostname, 100);
	PXgethostaddr(hostname, hostaddr);
	fprintf(stderr, "new DISPLAY = %s:%d or %s:%d\n",
		hostname, port-XPORT, hostaddr, port-XPORT
	);

	VXLocalPort = port;

	sprintf(buf, "DISPLAY=%s:%d", hostname, port-XPORT);
	putenv(buf);

	/* set hangul font */
	if((han_font = getenv(ENV_HAN_FONT)) != NULL) {
		strcpy(VXHanFontName, han_font);
	} else {
		strcpy(VXHanFontName, DEF_HAN_FONT);
	}

	/* set hangul font type : wanseong or johab */
	tmp = han_font_type = getenv(ENV_HAN_FONT_TYPE);
	if(han_font_type == NULL) {
		VXHanFontType = VX_JOHAB;
		puts("HanFontType = Johab");
	} else {
		while(*tmp != '\0') {
			if('a' <= *tmp && *tmp <= 'z')
				*tmp = *tmp - 'a' + 'A';
			tmp++;
		}
		if(strncmp(han_font_type, "WANS", 4) == 0) {
			puts("HanFontType = Wanseong");
			VXHanFontType = VX_WANSEONG;
		} else {
			puts("HanFontType = Johab");
			VXHanFontType = VX_JOHAB;
		}
	}

	/* set ascii font */
	if((asc_font = getenv(ENV_ASC_FONT)) != NULL) {
		strcpy(VXASCFontName, asc_font);
	} else {
		strcpy(VXASCFontName, DEF_ASC_FONT);
	}

	VXCheckSystemByteOrder();

	if (argc > arg_cmd)
	{
		pid1 = fork();
		if (pid1 == 0)
		{
			execvp(argv[arg_cmd], argv+arg_cmd);
			exit(1);
		}
#ifdef SIGCHLD
		signal(SIGCHLD, sig_child);
		VXInitChannel();
		VXMainLoop();
		exit(0);
#else
		pid2 = fork();
		if (pid2 == 0)
		{
			VXInitChannel();
			VXMainLoop();
			exit(1);
		}
		wait(NULL);
		kill(pid1,SIGTERM);
		kill(pid2,SIGTERM);
		exit(0);
#endif
	}
	else
	{
		VXInitChannel();
		VXMainLoop();
	}
	return 0;
}
