/*    %W% %G%
================================================================================
  Description :
  Author      : Ryu, Byeong-Soon
================================================================================
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include "socket.h"

/*	SockInitializeServer
--------------------------------------------------------------------------------
  Description : Initialize a server.
  Arguments   : 
  Return      : server's socket descriptor
--------------------------------------------------------------------------------
*/
int
SockInitializeServer(int port)
{
	int  s;
	char  *opt;
	char  hostname[MAX_HOSTNAME+1];
	struct sockaddr_in  sa;
	struct hostent  *hp;

	memset(&sa, 0, sizeof(struct sockaddr_in));
	PXgethostname(hostname, MAX_HOSTNAME);
	/*
	gethostname(hostname, MAX_HOSTNAME);
	*/

	hp = gethostbyname(hostname);
	if(hp == NULL) {
		perror("gethostbyname");
		return -1;
	}
	sa.sin_family = hp->h_addrtype;
	sa.sin_port = htons(port);

	if((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("Init-socket");
		return -1;
	}

	if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&opt,sizeof(opt)) < 0) {
		perror("sersockopt REUSEADDR");
		return -1;
	}

	if(bind(s, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
		perror("bind");
		close(s);
		return -1;
	}
	listen(s, 5);
	return(s);
}

/*	SockInitializeClient
--------------------------------------------------------------------------------
  Description : Initialize a client.
  Arguments   : 
  Return      : client's socket descriptor
--------------------------------------------------------------------------------
*/
int
SockInitializeClient(char *hostname, int port)
{
	int  sockfd;
	struct sockaddr_in  serv_addr;
	struct hostent  *hp;

	memset((char *) &serv_addr, 0, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);

	hp = gethostbyname(hostname);
	if(hp == NULL) {
		perror("gethostbyname");
		return -1;
	}
	memcpy(&(serv_addr.sin_addr.s_addr), hp->h_addr, hp->h_length);

	if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		return -1;
	}

	if(connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
		perror("connect");
		return -1;
	}

	return sockfd;
}

/*	SockAccept
--------------------------------------------------------------------------------
  Description : 
  Arguments   : 
  Return      : 
--------------------------------------------------------------------------------
*/
void
SockAccept(int sid, SockHandler doit, int do_fork)
{
	int  s, len, child_pid;
	struct sockaddr_in  from;

	len = sizeof(from);
	while(1) {
		s = accept(sid, (struct sockaddr *) &from, &len);
		if(s < 0) {
			perror("accept");
			continue;
		}

		if(do_fork == DO_FORK) {
			if((child_pid = fork()) < 0) {
				perror("fork");
				continue;
			}
			if(child_pid == 0) {
				close(sid);
				doit(s, &from);
				exit(0);
			}
		} else {
			doit(s, &from);
		}
		close(s);
	}
}

/*	SockReadN
--------------------------------------------------------------------------------
  Description : Read "n" bytes from a descriptor.
  Arguments   : 
  Return      : 
--------------------------------------------------------------------------------
*/
int
SockReadN(int fd, char *ptr, int nbytes)
{
	int  nleft, nread;

	nleft = nbytes;
	while(nleft > 0) {
		nread = read(fd, ptr, nleft);
		if(nread < 0)
			return nread;
		else if(nread == 0)			/* EOF */
			break;

		nleft -= nread;
		ptr += nread;
	}
	return(nbytes - nleft);
}

/*	SockWriteN
--------------------------------------------------------------------------------
  Description : Write "n" bytes to a descriptor.
  Arguments   : 
  Return      : 
--------------------------------------------------------------------------------
*/
int
SockWriteN(int fd, char *ptr, int nbytes)
{
	int  nleft, nwritten;

	nleft = nbytes;
	while(nleft > 0) {
		nwritten = write(fd, ptr, nleft);
		if(nwritten <= 0)
			return(nwritten);

		nleft -= nwritten;
		ptr += nwritten;
	}
	return(nbytes - nleft);
}

/*	SockReadLine
--------------------------------------------------------------------------------
  Description : Read a line from a descriptor.
  Arguments   : 
  Return      : 
--------------------------------------------------------------------------------
*/
int
SockReadLine(int fd, char *ptr, int maxlen)
{
	int  n, rc;
	char  c;

	for(n = 1; n < maxlen; n++) {
		if((rc = read(fd, &c, 1)) == 1) {
			*ptr++ = c;
			if(c == '\n')
				break;
		} else if(rc == 0) {
			if(n == 1)
				return(0);		/* EOF, no data read */
			else
				break;			/* EOF, some data was read */
		} else
			return(-1);
	}

	*ptr = 0;
	return(n);
}
