/*    %W% %G%
================================================================================
 Description :
 Author      : Ryu, Byeong-Soon
================================================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "pxhan.h"

/*#define TRACE_ON*/
#include "debug.h"

#define HAN_FONT			(ptr->res_base | 0xfffe)
#define ASC_FONT			(ptr->res_base | 0xfffd)

#define VX_MODE_LEN_STR		0
#define VX_MODE_DELTA		1
#define VX_MODE_FONT3		2
#define VX_MODE_FONT2		3
#define VX_MODE_FONT1		4
#define VX_MODE_FONT0		5
#define VX_MODE_STRING8		6

void
VXPutChar8(Channel *ptr, int i),
VXDoneFilterText(Channel *ptr),
VXNewFont(Channel *ptr, long font),
VXPutHangul(Channel *ptr);

void
VXFilterImageText8(Channel *xptr)
{
	register Channel *ptr;
	register unsigned long i, len;
	register char  *temp;
	long  gc;

	ptr = xptr;
	TRACE("begin VXFilterImageText8\n");
	len = ptr->c_str_req_len = (unsigned char)ptr->ci_buffer[1];
	temp = ptr->ci_buffer + 16;
	ptr->input_temp = 0;
	for(i = 0; i < len; i++) {
		if(*temp++ & 0x80)
			break;
	}

	/* if it doesn't contain hangul */
	if(len < 2 || i == len) {
		VXMemCopyClient(ptr, ptr->c_req_len);
		TRACE("there is no hangul\n");
		return;
	}

	if (ptr->ctemp == NULL)
	{
		if ((ptr->ctemp = (char*) malloc (MAX_BUFFER)) == NULL)
		{
			TRACE("VXFilterImageText8: Out of memory.. crashing...:(\n");
			exit(2);
		}
		ptr->c_temp_size = MAX_BUFFER;
	}
	memcpy(ptr->ctemp, ptr->ci_buffer, 16);

	ptr->ctemp[0] = OP_POLY_TEXT16;
	ptr->hangul = 0;
	ptr->c_len_field = -1;
	ptr->c_str_len = 0;

	gc = VXMemLong(ptr, ptr->ci_buffer, 8);
	ptr->c_font = VXGetFontGC(ptr, gc);
	if(ptr->c_font == 0)
		ptr->c_font = ASC_FONT;

	/* first character is not hangul */
	if((ptr->ci_buffer[16] & ptr->ci_buffer[17] & 0x80) == 0) {
		ptr->c_len_field = 16;
		ptr->ctemp[17] = 0;			/* delta */
		ptr->c_opos = 18;
	} else
		ptr->c_opos = 16;

	for(i = 16, len += 16; i < len; i++) {
		if(ptr->c_temp_size < (ptr->c_opos+50)) {
			TRACE("VXFilterImageText8: ctemp overlow..enlarging..\n");
			while (ptr->c_temp_size < ptr->c_opos + 50)
				ptr->c_temp_size += MAX_BUFFER;
			temp = (char *) realloc(ptr->ctemp, ptr->c_temp_size);
			if (temp == NULL)
			{
				TRACE("VXFilterImageText8:Out of mem. aborting..\n");
				exit(1);
			}
			TRACE1("VXFilterImageText8: ctemp enlarged to %d bytes.\n", ptr->c_temp_size);
			ptr->ctemp = temp;
		}
		VXPutChar8(ptr, i);
	}

	VXDoneFilterText(ptr);

	/* remove processed input data */
	ptr->ci_buff_size -= ptr->c_req_len;
	if(ptr->ci_buff_size > 0) {
		memmove(ptr->ci_buffer,
			ptr->ci_buffer+ptr->c_req_len,
			ptr->ci_buff_size);
	}
	TRACE("end VXFilterImageText8\n");
}

void
VXFilterPolyText8(Channel *xptr)
{
	register Channel *ptr;
	register unsigned long i, j, len;
	register char  *temp;
	long  gc;

	ptr = xptr;
	TRACE("begin VXFilterPolyText8");
	for(i = 0, j = ptr->c_ipos, len = ptr->ci_buff_size; i < len; i++, j++) {
		if(j < 16) {	/* header */
			switch(j) {
			case 0:		/* start */
				if (ptr->ctemp == NULL)
				{
					ptr->ctemp = (char*)malloc(MAX_BUFFER);
					if (ptr->ctemp == NULL)
					{
						TRACE("VXFilterPolyText8: Out of buf..crashed.\n");
						exit(3);
					}
					ptr->c_temp_size = MAX_BUFFER;
				}
				ptr->ctemp[0] = OP_POLY_TEXT16;
				ptr->c_opos = 1;
				ptr->hangul = 0;
				ptr->c_str_len = 0;
				TRACE1("req len = %d\n", ptr->c_req_len);
				break;
			case 11:	/* gc */
				ptr->ctemp[ptr->c_opos++] = ptr->ci_buffer[i];
				gc = VXMemLong(ptr, ptr->ctemp, 8);
				TRACE1("gc = %x\n", gc);
				ptr->c_font = VXGetFontGC(ptr, gc);
				if(ptr->c_font == 0)
					ptr->c_font = ASC_FONT;
				break;
			default:
				ptr->ctemp[ptr->c_opos++] = ptr->ci_buffer[i];
				if(j == 15)
					ptr->c_mode = VX_MODE_LEN_STR;
				break;
			}
		} else {	/* data area */
			if(ptr->c_temp_size < (ptr->c_opos+50)) {
				TRACE("VXFilterPolyText8: ctemp overlow..enlarging..\n");
				while (ptr->c_temp_size < ptr->c_opos + 50)
					ptr->c_temp_size += MAX_BUFFER;
				temp = (char *) realloc(ptr->ctemp, ptr->c_temp_size);
				if (temp == NULL)
				{
					TRACE("ptr->c_temp_size: Out of mem. aborting..\n");
					exit(1);
				}
				TRACE1("VXFilterPolyText8: ctemp enlarged to %d bytes.\n", ptr->c_temp_size);
				ptr->ctemp = temp;
			}
			if(j == ptr->c_req_len)
				break;

			switch(ptr->c_mode) {
			case VX_MODE_LEN_STR:
				ptr->c_str_req_len = (unsigned char)ptr->ci_buffer[i];
				TRACE1("str_req_len = %d\n", ptr->c_str_req_len);
				if(j >= ptr->c_req_len-2 ||
					(j == ptr->c_req_len-3 && ptr->c_str_req_len == 0)
				) {
					break;
				} else if(ptr->c_str_req_len == 255) {
					ptr->c_mode = VX_MODE_FONT3;
					ptr->ctemp[ptr->c_opos++] = 255;
					ptr->c_len_field = -1;
				} else {
					ptr->c_len_field = ptr->c_opos++;
					ptr->ctemp[ptr->c_len_field] = 0;
					ptr->c_mode = VX_MODE_DELTA;
				}
				ptr->c_str_len = 0;
				break;
			case VX_MODE_DELTA:
				ptr->ctemp[ptr->c_opos++] = ptr->ci_buffer[i];
				if(ptr->c_str_req_len == 0)
					ptr->c_mode = VX_MODE_LEN_STR;
				else
					ptr->c_mode = VX_MODE_STRING8;
				break;
			case VX_MODE_FONT3:
				ptr->ctemp[ptr->c_opos++] = ptr->ci_buffer[i];
				ptr->c_mode = VX_MODE_FONT2;
				break;
			case VX_MODE_FONT2:
				ptr->ctemp[ptr->c_opos++] = ptr->ci_buffer[i];
				ptr->c_mode = VX_MODE_FONT1;
				break;
			case VX_MODE_FONT1:
				ptr->ctemp[ptr->c_opos++] = ptr->ci_buffer[i];
				ptr->c_mode = VX_MODE_FONT0;
				break;
			case VX_MODE_FONT0:
				ptr->ctemp[ptr->c_opos++] = ptr->ci_buffer[i];
				ptr->c_font = VXMemMSBLong(ptr->ctemp, ptr->c_opos-4);
				ptr->c_mode = VX_MODE_LEN_STR;
				break;
			case VX_MODE_STRING8:
				VXPutChar8(ptr, i);
				if(ptr->c_str_req_len == 0) {
					ptr->c_mode = VX_MODE_LEN_STR;
				}
				break;
			}
		}
	}

	ptr->c_ipos = j;
	if(j == ptr->c_req_len)
		VXDoneFilterText(ptr);
	/* remove processed input data */
	TRACE1("remove input: input_temp = %d\n", ptr->input_temp);
	ptr->ci_buff_size -= i;
	if(ptr->ci_buff_size != 0)
		memmove(ptr->ci_buffer, ptr->ci_buffer+i, ptr->ci_buff_size);
	
	TRACE("VXFilterPolyText8\n");
}

void
VXPutChar8(Channel *xptr, int i)
{
	register Channel *ptr;
	char  ch, new_ch;

	ptr = xptr;
	new_ch = ptr->ci_buffer[i];

	if(new_ch & 0x80) {
		TRACE2("HCH(%d) = '%x'\n", ptr->c_ipos, new_ch);
		ptr->hangul++;
		if(ptr->hangul == 2) {
			ch = ptr->ctemp[--(ptr->c_opos)];
			VXNewFont(ptr, HAN_FONT);
			ptr->ctemp[ptr->c_opos++] = ch;
		}
		ptr->ctemp[ptr->c_opos++] = new_ch;
		if((ptr->hangul % 2) == 0)
			VXPutHangul(ptr);
	} else {
		TRACE2("CH(%d) = '%c'\n", ptr->c_ipos, ptr->ci_buffer[i]);
		if(ptr->hangul != 0) {
			if((ptr->hangul%2) == 1)
				ch = ptr->ctemp[--(ptr->c_opos)];
			if(ptr->hangul >= 2)
				VXNewFont(ptr,ptr->c_font);
			if((ptr->hangul%2) == 1)
			{
				ptr->ctemp[ptr->c_opos++] = 0;
				ptr->ctemp[ptr->c_opos++] = ch;
				ptr->c_str_len++;

			}
			ptr->hangul = 0;
		}
		ptr->ctemp[ptr->c_opos++] = 0;
		ptr->ctemp[ptr->c_opos++] = new_ch;
		ptr->c_str_len++;
	}

	ptr->c_str_req_len--;
	if(ptr->c_str_req_len == 0) {
		if((ptr->hangul%2) == 1) {
			TRACE("remain hangul\n");
			ch = ptr->ctemp[--(ptr->c_opos)];
			if(ptr->hangul != 1)
				VXNewFont(ptr,ptr->c_font);
			ptr->ctemp[ptr->c_opos++] = 0;
			ptr->ctemp[ptr->c_opos++] = ch;
			ptr->c_str_len++;
			ptr->hangul = 0;
		}
		TRACE2("len_field = %d, len = %d\n", ptr->c_len_field, ptr->c_str_len);
		ptr->ctemp[ptr->c_len_field] = (char)ptr->c_str_len;
	}
}

void
VXDoneFilterText(Channel *xptr)
{
	register Channel *ptr;
	register unsigned long i;
	register char *temp;

	ptr = xptr;
	if(ptr->hangul != 0) {
		VXNewFont(ptr, ptr->c_font);
		ptr->ctemp[ptr->c_len_field] = 0;
	}
	ptr->input_temp = 1;
	for (i=ptr->c_opos, temp=(ptr->ctemp+i); i%4; i++, temp++)
		*temp = 0;
	ptr->c_opos = i;
	if (ptr->byte_order == OP_MSB_FIRST)
		VXSetMemMSBShort(ptr->ctemp, 2, (short)(ptr->c_opos/4));
	else
		VXSetMemLSBShort(ptr->ctemp, 2, (short)(ptr->c_opos/4));
}

void
VXNewFont(Channel *xptr, long font)
{
	register Channel *ptr;

	ptr = xptr;
	TRACE2("FONT = %x,   opos = %d\n", font, ptr->c_opos);
	if(ptr->c_len_field != -1) {
		TRACE2("LEN_FIELD = %d, len = %d\n", ptr->c_len_field, ptr->c_str_len);
		ptr->ctemp[ptr->c_len_field] = (char)ptr->c_str_len;
	}
	ptr->ctemp[ptr->c_opos++] = 255;		/* font shift */
	VXSetMemMSBLong(ptr->ctemp, ptr->c_opos, font);
	ptr->c_opos += 4;
	ptr->c_len_field = ptr->c_opos++;
	ptr->ctemp[ptr->c_opos++] = 0;		/* delta = 0 */
	ptr->c_str_len = 0;
	TRACE1("opos = %d\n", ptr->c_opos);
}

void
VXPutHangul(Channel *xptr)
{
	register Channel *ptr;
	unsigned char  wan[2], han3[3];
	int  n;
	char  des[3][2];

	ptr = xptr;

	if(VXHanFontType == VX_JOHAB) {
		ptr->c_opos -= 2;
		wan[0] = ptr->ctemp[ptr->c_opos];
		wan[1] = ptr->ctemp[ptr->c_opos+1];

		convert_ks_to_3(wan[0], wan[1], han3);
		TRACE2("HAN: %c%c\n", wan[0], wan[1]);
		TRACE3("H3: %d, %d, %d\n", han3[0], han3[1], han3[2]);
		n = convert_3_to_display((int)han3[0], (int)han3[1], (int)han3[2], des);
		TRACE4("DISP: n = %d, (%d,%d,%d)\n",
			n, des[0][0], des[1][0], des[2][0]
		);
		TRACE4("DISP: n = %d, (%d,%d,%d)\n",
			n, des[0][1], des[1][1], des[2][1]
		);

		memcpy((char*)(ptr->ctemp+ptr->c_opos), (char*)des, 2*n);
		ptr->c_opos += 2*n;
		ptr->c_str_len += n;
	} else {
		ptr->ctemp[ptr->c_opos-2] &= 0x7f;
		ptr->ctemp[ptr->c_opos-1] &= 0x7f;
		ptr->c_str_len++;
	}
}

void
VXInitHangulFont(Channel *xptr)
{
	register Channel *ptr;
	unsigned long  slen, len, pad;

	ptr = xptr;
	slen = strlen(VXHanFontName);
	pad = 4 - (slen % 4);
	if(pad == 4)
		pad = 0;
	len = slen + pad + 12;
	ptr->so_buffer[ptr->so_buff_size++] = OP_OPEN_FONT;
	ptr->so_buff_size++;
	if (ptr->byte_order == OP_MSB_FIRST)
		VXSetMemMSBShort(ptr->so_buffer, ptr->so_buff_size, len/4);
	else
		VXSetMemLSBShort(ptr->so_buffer, ptr->so_buff_size, len/4);
	ptr->so_buff_size += 2;
	if (ptr->byte_order == OP_MSB_FIRST)
		VXSetMemMSBLong(ptr->so_buffer, ptr->so_buff_size, HAN_FONT);
	else
		VXSetMemLSBLong(ptr->so_buffer, ptr->so_buff_size, HAN_FONT);
	ptr->so_buff_size += 4;
	if (ptr->byte_order == OP_MSB_FIRST)
		VXSetMemMSBShort(ptr->so_buffer, ptr->so_buff_size, slen);
	else
		VXSetMemLSBShort(ptr->so_buffer, ptr->so_buff_size, slen);
	ptr->so_buff_size += 4;
	memcpy(ptr->so_buffer+ptr->so_buff_size, VXHanFontName, slen);
	ptr->so_buff_size += slen + pad;

	/* below is temporary */
	slen = strlen(VXASCFontName);
	pad = 4 - (slen % 4);
	if(pad == 4)
		pad = 0;
	len = slen + pad + 12;
	ptr->so_buffer[ptr->so_buff_size++] = OP_OPEN_FONT;
	ptr->so_buff_size++;
	if (ptr->byte_order == OP_MSB_FIRST)
		VXSetMemMSBShort(ptr->so_buffer, ptr->so_buff_size, len/4);
	else
		VXSetMemLSBShort(ptr->so_buffer, ptr->so_buff_size, len/4);
	ptr->so_buff_size += 2;
	if (ptr->byte_order == OP_MSB_FIRST)
		VXSetMemMSBLong(ptr->so_buffer, ptr->so_buff_size, ASC_FONT);
	else
		VXSetMemLSBLong(ptr->so_buffer, ptr->so_buff_size, ASC_FONT);
	ptr->so_buff_size += 4;
	if (ptr->byte_order == OP_MSB_FIRST)
		VXSetMemMSBShort(ptr->so_buffer, ptr->so_buff_size, slen);
	else
		VXSetMemLSBShort(ptr->so_buffer, ptr->so_buff_size, slen);
	ptr->so_buff_size += 4;
	memcpy(ptr->so_buffer+ptr->so_buff_size, VXASCFontName, slen);
	ptr->so_buff_size += slen + pad;
	ptr->c_font = ASC_FONT;
}
